/* file: svm_predict_oneapi_impl.i */
/*******************************************************************************
* Copyright 2020 Intel Corporation
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*******************************************************************************/

/*
//++
//  SVM prediction algorithm implementation
//--
*/

#ifndef __SVM_PREDICT_ONEAPI_IMPL_I__
#define __SVM_PREDICT_ONEAPI_IMPL_I__

#include "src/sycl/blas_gpu.h"
#include "src/externals/service_ittnotify.h"
#include "data_management/data/internal/numeric_table_sycl_homogen.h"
#include "src/algorithms/svm/oneapi/svm_helper_oneapi.h"

DAAL_ITTNOTIFY_DOMAIN(svm_predict.default.batch);

namespace daal
{
namespace algorithms
{
namespace svm
{
namespace prediction
{
namespace internal
{
using namespace daal::internal;
using namespace daal::services::internal::sycl;
using daal::data_management::internal::SyclHomogenNumericTable;

template <typename algorithmFPType>
services::Status SVMPredictImplOneAPI<defaultDense, algorithmFPType>::compute(const NumericTablePtr & xTable, Model * model, NumericTable & result,
                                                                              const svm::Parameter * par)
{
    services::Status status;
    auto & context = services::internal::getDefaultContext();

    const size_t nVectors  = xTable->getNumberOfRows();
    const size_t nFeatures = xTable->getNumberOfColumns();

    DAAL_ASSERT(result.getNumberOfRows() == nVectors)
    DAAL_ASSERT(result.getNumberOfColumns() == 1)

    BlockDescriptor<algorithmFPType> resultBlock;
    DAAL_CHECK_STATUS(status, result.getBlockOfRows(0, nVectors, ReadWriteMode::writeOnly, resultBlock));
    auto distanceBuff = resultBlock.getBuffer();

    kernel_function::KernelIfacePtr kernel = par->kernel->clone();

    auto svCoeffTable = model->getClassificationCoefficients();
    const size_t nSV  = svCoeffTable->getNumberOfRows();

    if (nSV == 0)
    {
        context.fill(distanceBuff, 0.0, status);
        return status;
    }

    BlockDescriptor<algorithmFPType> svCoeffBlock;
    DAAL_CHECK_STATUS(status, svCoeffTable->getBlockOfRows(0, nSV, ReadWriteMode::readOnly, svCoeffBlock));
    auto svCoeffBuff = svCoeffBlock.getBuffer();

    const algorithmFPType bias(model->getBias());
    context.fill(distanceBuff, double(bias), status);
    DAAL_CHECK_STATUS_VAR(status);

    auto svTable = model->getSupportVectors();

    const size_t nRowsPerBlock = 1024;
    const size_t nBlocks       = nVectors / nRowsPerBlock + !!(nVectors % nRowsPerBlock);

    DAAL_OVERFLOW_CHECK_BY_MULTIPLICATION(size_t, nRowsPerBlock, nSV);
    auto kernelResU = context.allocate(TypeIds::id<algorithmFPType>(), nRowsPerBlock * nSV, status);
    DAAL_CHECK_STATUS_VAR(status);

    auto kernelResBuff = kernelResU.template get<algorithmFPType>();

    kernel_function::ResultPtr shRes(new kernel_function::Result());
    DAAL_CHECK_MALLOC(shRes)

    for (size_t iBlock = 0; iBlock < nBlocks; ++iBlock)
    {
        const size_t startRow          = iBlock * nRowsPerBlock;
        const size_t nRowsPerBlockReal = (iBlock != nBlocks - 1) ? nRowsPerBlock : nVectors - iBlock * nRowsPerBlock;
        NumericTablePtr kernelResNT = SyclHomogenNumericTable<algorithmFPType>::create(kernelResBuff.getSubBuffer(0, nRowsPerBlockReal * nSV, status),
                                                                                       nSV, nRowsPerBlockReal, &status);
        DAAL_CHECK_STATUS_VAR(status);

        BlockDescriptor<algorithmFPType> xBlock;
        DAAL_CHECK_STATUS(status, xTable->getBlockOfRows(startRow, nRowsPerBlockReal, ReadWriteMode::readOnly, xBlock));
        const services::internal::Buffer<algorithmFPType> xBuf = xBlock.getBuffer();

        NumericTablePtr xBlockNT = SyclHomogenNumericTable<algorithmFPType>::create(xBuf, nFeatures, nRowsPerBlockReal, &status);
        DAAL_CHECK_STATUS_VAR(status);

        shRes->set(kernel_function::values, kernelResNT);
        kernel->setResult(shRes);

        kernel->getInput()->set(kernel_function::X, xBlockNT);
        kernel->getInput()->set(kernel_function::Y, svTable);
        kernel->getParameter()->computationMode = kernel_function::matrixMatrix;
        DAAL_CHECK(kernel->computeNoThrow(), services::ErrorSVMPredictKernerFunctionCall);

        {
            DAAL_ITTNOTIFY_SCOPED_TASK(gemm);
            DAAL_CHECK_STATUS(status, BlasGpu<algorithmFPType>::xgemm(math::Layout::RowMajor, math::Transpose::NoTrans, math::Transpose::NoTrans,
                                                                      nRowsPerBlockReal, 1, nSV, algorithmFPType(1.0), kernelResBuff, nSV, 0,
                                                                      svCoeffBuff, 1, 0, algorithmFPType(1.0), distanceBuff, 1, startRow));
        }

        DAAL_CHECK_STATUS(status, xTable->releaseBlockOfRows(xBlock));
    }

    DAAL_CHECK_STATUS(status, result.releaseBlockOfRows(resultBlock));
    DAAL_CHECK_STATUS(status, svCoeffTable->releaseBlockOfRows(svCoeffBlock));

    return status;
}

} // namespace internal
} // namespace prediction
} // namespace svm
} // namespace algorithms
} // namespace daal

#endif
